// ------------------------------------------------------------------------
// Copyright (C) 1996, ITS Information Technology Solutions GmbH.
//
// This file is part of ITSEDIPP, the IT Solutions EDI Pseudo Printer.
//
// ITSEDIPP is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// ITSEDIPP is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with ITSEDIPP; see the file COPYING.  If not, write to the
// Free Software Foundation, Inc, 675 Mass Ave, Cambridge, MA 02139, USA.
// 
// Support for ITSEDIPP can be obtained from (as of November 1996):
//	"IT Solutions"
//	ITS Information Technology Solutions GmbH
//	Balanstr. 69 / EG
//	D - 81541 Munich
//	Germany
//	Phone:	+49(89)400424
//	E-Mail:	100335.120@compuserve.com
// ------------------------------------------------------------------------
//
// HEADER SEPARATOR

#include	"itspmdde.h"

DWORD	dwIDInst = 0;
HSZ	hszService = NULL;

VOID INTERN
DDEStart(
	HWND	hWnd
	)
{
	UINT	iStat;

	IFDBG(debug("DDEStart", "called", NULL, IPMD_NOARG));
	SendMessage(hWnd, IPMD_LOGEVENT, IPMD_DDESTART, (LPARAM) IPMD_SERVICE);

	if (dwIDInst != 0)
	{
		error("DDEStart",
			"ID instance already initialized", NULL, dwIDInst);
		return;
	}

	// initialize DDEML
	iStat = DdeInitialize(&dwIDInst, DDECallback,
					APPCLASS_STANDARD |
					APPCMD_FILTERINITS |
					CBF_FAIL_ADVISES |
					CBF_FAIL_REQUESTS |
					CBF_FAIL_SELFCONNECTIONS |
					CBF_SKIP_REGISTRATIONS |
					CBF_SKIP_UNREGISTRATIONS |
					0 , 0);
	if (iStat != DMLERR_NO_ERROR)
	{
		error("DDEStart",
			"cannot DDE initialize", NULL, (DWORD) iStat);
		dwIDInst = 0;
		return;
	}
	IFDBG(debug("DDEStart", "ID instance", NULL, dwIDInst));

	// get handle for service name
	hszService = DdeCreateStringHandle(dwIDInst, IPMD_SERVICE, 0);
	if (hszService == NULL)
	{
		error("DDEStart",
			"cannot create DDE string handle", NULL,
					(DWORD) DdeGetLastError(dwIDInst));
		DdeUninitialize(dwIDInst);
		dwIDInst = 0;
		return;
	}
	IFDBG(debug("DDEStart", "service string handle", NULL, (DWORD) hszService));

	// register service name with DDEML
	if (DdeNameService(dwIDInst, hszService, NULL, DNS_REGISTER) == 0)
	{
		error("DDEStart",
			"cannot register DDE service", NULL,
					(DWORD) DdeGetLastError(dwIDInst));
		DdeFreeStringHandle(dwIDInst, hszService);
		DdeUninitialize(dwIDInst);
		dwIDInst = 0;
		return;
	}

	IFDBG(debug("DDEStart", "DDE service established", IPMD_SERVICE, IPMD_NOARG));
}

VOID INTERN
DDEStop(
	HWND	hWnd
	)
{
	IFDBG(debug("DDEStop", "called", NULL, IPMD_NOARG));
	SendMessage(hWnd, IPMD_LOGEVENT, IPMD_DDESTOP, (LPARAM) IPMD_SERVICE);

	if (dwIDInst == 0)
	{
		error("DDEStop",
			"ID instance not initialized", NULL, IPMD_NOARG);
		return;
	}

	// un-register all service names with DDEML
	DdeNameService(dwIDInst, NULL, NULL, DNS_UNREGISTER);

	// release service name handle
	if (hszService != NULL)
	{
		DdeFreeStringHandle(dwIDInst, hszService);
		hszService = NULL;
	}

	// un-initialize DDEML
	DdeUninitialize(dwIDInst);
	dwIDInst = 0;

	IFDBG(debug("DDEStop", "DDE service terminated", IPMD_SERVICE, IPMD_NOARG));
}

HDDEDATA EXPORT
DDECallback(
	UINT		iType,
	UINT		iFmt,
	HCONV		hConv,
	HSZ		hsz1,
	HSZ		hsz2,
	HDDEDATA	hData,
	DWORD		dwData1,
	DWORD		dwData2
	)
{
	CHAR	szInService[16];
	CHAR	szInTopic[16];
	CHAR	szInItem[32];
	LPSTR	lpData;
	DWORD	dwLen;
	BOOL	bExecOK;
	LPSTR	lpExec;
	BOOL	bPokeOK;
	LPSTR	lpPoke;

	IFDBG(debug("DDECallback", "message type", NULL, (DWORD) iType));

	switch (iType)
	{
	case XTYP_CONNECT:
		// hsz1, hsz2 == topic, service
		IFDBG(SendMessage(hDialog, IPMD_LOGEVENT, IPMD_DDECALLBACK, (LPARAM) "connect"));

		DdeQueryString(dwIDInst, hsz2, szInService, sizeof(szInService), 0);
		DdeQueryString(dwIDInst, hsz1, szInTopic, sizeof(szInTopic), 0);

		if (_fstricmp(szInService, IPMD_SERVICE) != 0)
		{
			error("DDECallback",
				"unrecognized service name", szInService, IPMD_NOARG);
			return((HDDEDATA) FALSE);
		}

		if (_fstricmp(szInTopic, IPMD_TOPIC) != 0)
		{
			error("DDECallback",
				"unrecognized topic name", szInTopic, IPMD_NOARG);
			return((HDDEDATA) FALSE);
		}

		return((HDDEDATA) TRUE);

	case XTYP_CONNECT_CONFIRM:
		// hConv == new conversation
		IFDBG(SendMessage(hDialog, IPMD_LOGEVENT, IPMD_DDECALLBACK, (LPARAM) "connect confirm"));

		if (got_CONNECTION(hConv) == FALSE)
		{
			error("DDECallback",
				"cannot establish connection", NULL, IPMD_NOARG);
		}

		break;

	case XTYP_DISCONNECT:
		// hConv == ending conversation
		IFDBG(SendMessage(hDialog, IPMD_LOGEVENT, IPMD_DDECALLBACK, (LPARAM) "disconnect"));

		if (got_DISCONNECTION(hConv) == FALSE)
		{
			error("DDECallback",
				"cannot disconnect", NULL, IPMD_NOARG);
		}

		break;

	case XTYP_EXECUTE:
		// hConv, hsz1, hData == conversation, topic, command
		IFDBG(SendMessage(hDialog, IPMD_LOGEVENT, IPMD_DDECALLBACK, (LPARAM) "execute"));

		lpData = DdeAccessData(hData, &dwLen);
		if (lpData == NULL)
		{
			error("DDECallback",
				"cannot access execute data", NULL,
					(DWORD) DdeGetLastError(dwIDInst));
			return((HDDEDATA) DDE_FNOTPROCESSED);
		}

		// string_value allocates copy of lpData and nul-terminates it
		lpExec = string_value(lpData, (WORD) dwLen);

		DdeUnaccessData(hData);

		if (lpExec == NULL)
		{
			error("DDECallback",
				"cannot copy execute data", NULL,
					IPMD_NOARG);
			return((HDDEDATA) DDE_FNOTPROCESSED);
		}

		// remove trailing <CR>s and <LF>s, if present
		chop_crlf(lpExec);

		if (_fstricmp(lpExec, "NEWMESSAGE") == 0)
			bExecOK = got_NEWMESSAGE(hConv);
		else if (_fstricmp(lpExec, "SETDEFAULTS") == 0)
			bExecOK = got_SETDEFAULTS(hConv);
		else if (_fstricmp(lpExec, "SENDMESSAGE") == 0)
			bExecOK = got_SENDMESSAGE(hConv);
		else
		{
			error("DDECallback",
				"unrecognized command", lpExec, IPMD_NOARG);
			_ffree(lpExec);
			return((HDDEDATA) DDE_FNOTPROCESSED);
		}
		
		if (bExecOK == FALSE)
		{
			error("DDECallback",
				"error executing command", lpExec, IPMD_NOARG);
			_ffree(lpExec);
			return((HDDEDATA) DDE_FNOTPROCESSED);
		}

		_ffree(lpExec);
		return((HDDEDATA) DDE_FACK);

	case XTYP_POKE:
		// hConv, hsz1, hsz2, hData == conversation, topic, item, value
		IFDBG(SendMessage(hDialog, IPMD_LOGEVENT, IPMD_DDECALLBACK, (LPARAM) "poke"));

		if (iFmt != CF_TEXT)
		{
			IFDBG(debug("DDECallback", "unknown poke data format", NULL, (DWORD) iFmt));
			return((HDDEDATA) DDE_FNOTPROCESSED);
		}

		DdeQueryString(dwIDInst, hsz2, szInItem, sizeof(szInItem), 0);

		lpData = DdeAccessData(hData, &dwLen);
		if (lpData == NULL)
		{
			error("DDECallback",
				"cannot access poke data", NULL,
					(DWORD) DdeGetLastError(dwIDInst));
			return((HDDEDATA) DDE_FNOTPROCESSED);
		}

		// string_value allocates copy of lpData and nul-terminates it
		lpPoke = string_value(lpData, (WORD) dwLen);

		DdeUnaccessData(hData);

		if (lpPoke == NULL)
		{
			error("DDECallback",
				"cannot copy poke data", NULL,
					IPMD_NOARG);
			return((HDDEDATA) DDE_FNOTPROCESSED);
		}

		// do not remove trailing <CR><LF> sequence from lpPoke;
		// called function will take care of that, if necessary

		if (_fstricmp(szInItem, "TO") == 0)
			bPokeOK = got_TO(hConv, lpPoke);
		else if (_fstricmp(szInItem, "CC") == 0)
			bPokeOK = got_CC(hConv, lpPoke);
		else if (_fstricmp(szInItem, "BCC") == 0)
			bPokeOK = got_BCC(hConv, lpPoke);
		else if (_fstricmp(szInItem, "SUBJECT") == 0)
			bPokeOK = got_SUBJECT(hConv, lpPoke);
		else if (_fstricmp(szInItem, "COPYSELF") == 0)
			bPokeOK = got_COPYSELF(hConv, lpPoke);
		else if (_fstricmp(szInItem, "CONFIRMREADING") == 0)
			bPokeOK = got_CONFIRMREADING(hConv, lpPoke);
		else if (_fstricmp(szInItem, "CONFIRMDELIVERY") == 0)
			bPokeOK = got_CONFIRMDELIVERY(hConv, lpPoke);
		else if (_fstricmp(szInItem, "URGENT") == 0)
			bPokeOK = got_URGENT(hConv, lpPoke);
		else if (_fstricmp(szInItem, "MIME") == 0)
			bPokeOK = got_MIME(hConv, lpPoke);
		else if (_fstricmp(szInItem, "8BITDATA") == 0)
			bPokeOK = got_8BITDATA(hConv, lpPoke);
		else if (_fstricmp(szInItem, "REPLYTO") == 0)
			bPokeOK = got_REPLYTO(hConv, lpPoke);
		else if (_fstricmp(szInItem, "ATTACHFILE") == 0)
			bPokeOK = got_ATTACHFILE(hConv, lpPoke);
		else if (_fstricmp(szInItem, "ATTACHTYPE") == 0)
			bPokeOK = got_ATTACHTYPE(hConv, lpPoke);
		else if (_fstricmp(szInItem, "ATTACHENCODING") == 0)
			bPokeOK = got_ATTACHENCODING(hConv, lpPoke);
		else if (_fstricmp(szInItem, "ENCRYPTIONKEY") == 0)
			bPokeOK = got_ENCRYPTIONKEY(hConv, lpPoke);
		else if (_fstricmp(szInItem, "NOSIG") == 0)
			bPokeOK = got_NOSIG(hConv, lpPoke);
		else if (_fstricmp(szInItem, "MESSAGEFILE") == 0)
			bPokeOK = got_MESSAGEFILE(hConv, lpPoke);
		else if (_fstricmp(szInItem, "MESSAGETEXT") == 0)
			bPokeOK = got_MESSAGETEXT(hConv, lpPoke);
		else if (_fstricmp(szInItem, "EXPIRATION") == 0)
			bPokeOK = got_EXPIRATION(hConv, lpPoke);
		else if (_fstricmp(szInItem, "SETHEADERTYPE") == 0)
			bPokeOK = got_SETHEADERTYPE(hConv, lpPoke);
		else if (_fstricmp(szInItem, "SETHEADER") == 0)
			bPokeOK = got_SETHEADER(hConv, lpPoke);
		else
		{
			error("DDECallback",
				"unrecognized poke item", szInItem, IPMD_NOARG);
			_ffree(lpPoke);
			return((HDDEDATA) DDE_FNOTPROCESSED);
		}

		if (bPokeOK == FALSE)
		{
			error("DDECallback",
				"error processing poke", szInItem, IPMD_NOARG);
			_ffree(lpPoke);
			return((HDDEDATA) DDE_FNOTPROCESSED);
		}

		// if poke was completed ok, then called function has
		// freed or saved lpPoke and we do not have to do it here

		return((HDDEDATA) DDE_FACK);

	case XTYP_ERROR:
		// hConv == conversation with problem
		IFDBG(SendMessage(hDialog, IPMD_LOGEVENT, IPMD_DDECALLBACK, (LPARAM) "error"));

		error("DDECallback",
			"received DDE error message", NULL, IPMD_NOARG);
		got_DISCONNECTION(hConv);

		break;

	default:
		IFDBG(SendMessage(hDialog, IPMD_LOGEVENT, IPMD_DDECALLBACK, (LPARAM) "unrecognized"));

		error("DDECallback",
			"received unrecognized DDE message", NULL, (DWORD) iType);
		switch (iType & XCLASS_MASK)
		{
		case XCLASS_BOOL:
			return((HDDEDATA) FALSE);
		case XCLASS_DATA:
			return((HDDEDATA) NULL);
		case XCLASS_FLAGS:
			return((HDDEDATA) DDE_FNOTPROCESSED);
		case XCLASS_NOTIFICATION:
			;
		}
	}

	return((HDDEDATA) 0);
}
